`timescale 1ns/10ps

module model_1645RU5(D, A, nWE, nOE, nCE);
inout [7:0] D;
input [18:0] A;
input nWE, nOE, nCE;

//Read Cycle Parameters - 3.3V use - comment this section for 5V use
parameter Taad  = 30; // address access time
parameter Tanced = 30; // nCE     access time
parameter Tpzlnced =  3; // nCE to output low Z time
parameter Tplznced =  15; // nCE to output high Z time
parameter Tanoed  =  15; // nOE to output  time
parameter Tplznoed =  15; // nOE to output high Z time
parameter Tvad  =  5; // data hold from adr change time
      

//Write Cycle Parameters - 3.3V use - comment this section for 5V use
parameter Tsuanwelh  =  25; // adr valid to end of write time
parameter Twnce  =  25; // nCE to end of write time
parameter Tsuanwehl  =  0; // address set up time
parameter Twnwe  =  25; // write pulse width min
parameter Tsudnwelh  =  25; // data valid to end of writ time
parameter Tpzlnwed  =  5; // data low Z from end of writ time
parameter Tplznwed =  15; // nWE to output in high Z time

/*  uncomment for 5V use
//Read Cycle Parameters - 5V use
parameter Taad  = 20; // address access time
parameter Tanced = 20; // nCE     access time
parameter Tpzlnced =  3; // nCE to output low Z time
parameter Tplznced = 10; // nCE to output high Z time
parameter Tanoed  =  10; // nOE to output  time
parameter Tplznoed =  10; // nOE to output high Z time
parameter Tvad  =  5; // data hold from adr change time
      

//Write Cycle Parameters - 5V use
parameter Tsuanwelh  =  20; // adr valid to end of write time
parameter Twnce  =  20; // nCE to end of write time
parameter Tsuanwehl  =  0; // address set up time
parameter Twnwe  =  15; // write pulse width min
parameter Tsudnwelh  =  15; // data valid to end of writ time
parameter Tpzlnwed  =  5; // data act from end of writ time
parameter Tplznwed =  10; // nWE to output in high Z time
*/


reg [7:0] mem1[0:524287];

time adr_chng,da_chng,nWEfall,nWErise,nCEfall,nCErise;
time nOEfall,nOErise;

wire [7:0] data_in;
reg  [7:0] data_out;
reg  [7:0] temp1,temp2,temp3;
reg outen, out_en, in_en;


initial
  begin
       in_en <= 1'b1;
    if (nCE)
       out_en <= 1'b0;
     else out_en <= 1'b1;
  end
  
wire [18:0] addr;

assign addr = A;

// input/output control logic
//---------------------------
assign D   = out_en ? data_out : 'hzz;
assign data_in = in_en ? D : 'hzz;

// read access
//------------
always @(addr)
      if (nCE==0 & nWE==1) begin           //read

        #(Taad) data_out[7:0] <= mem1[addr];

      end
      
always @(addr)
  begin

              outen  <= 1'b0;
         #(Tvad) out_en <= outen;

//---------------------------------------------
      if (nCE==0 & nWE==1)                 //read
        begin
           if (nOE==0)
             begin
              outen <= 1'b1;
              out_en <= 1'b1;
             end
        end
//---------------------------------------------
     if (nCE==0 & nWE==0)                 //write
       begin
         if (nOE==0)
           begin
                outen <= 1'b0;
                out_en <= 1'b0;
                temp1 <= data_in;

                    #(Tsudnwelh) mem1[addr] <= temp1[7:0];

           end
         else
           begin
                outen <= 1'b0;
                out_en <= 1'b0;
                temp1 <= data_in;

                    #(Tsudnwelh-Tvad) mem1[addr] <= temp1[7:0];

           end

           data_out[7:0] <= mem1[addr];

       end
  end

always @(negedge nCE)
  begin
     nCEfall <= $time;

     if (nCEfall - adr_chng < Tsuanwehl)
         begin 
	 $display($time, "ns !!!ERROR!!!  Adr setup time is not enough Tsuanwehl");
	 $finish;
	 end

      if (nWE==1 & nOE==0)
               outen  <= 1'b1;
         #(Tpzlnced) out_en <= outen;

      if (nWE==1) begin

           #(Tanced-Tpzlnced) data_out[7:0] <= mem1[addr];

      end

      if (nWE==0)
       begin
               outen <= 1'b0;
               out_en <= 1'b0;
               temp2 <= data_in;

                 #(Tsudnwelh) mem1[addr] <= temp2[7:0];

       end
  end


always @(negedge nOE)
  begin
     nOEfall <= $time;
      

         data_out[7:0] <= mem1[addr];


      if (nWE==1 & nCE==0)
              outen  <= 1'b1;
         #(Tanoed) out_en <= outen;
  end

always @(posedge nOE)
  begin
     nOErise <= $time;

               outen  <= 1'b0;
         #(Tplznoed) out_en <= outen;
  end

// write to ram
//-------------
always @(negedge nWE)
  begin
     nWEfall <= $time;

     if (nWEfall - adr_chng < Tsuanwehl)
     	 begin
         $display($time, "ns !!!ERROR!!!  Address set-up to WE low is not enough");
	 $finish;
	 end

     if (nCE==0 & nOE==0)
       begin
               outen  <= 1'b0;
         #(Tplznwed) out_en <= outen;
                  temp3 <= data_in;

           #(Tsudnwelh) mem1[addr] <= temp3[7:0];



              data_out[7:0] <= mem1[addr];

       end

     if (nCE==0 & nOE==1)
       begin
               outen <= 1'b0;
               out_en <= 1'b0;
                  temp3 <= data_in;

           #(Tsudnwelh) mem1[addr] <= temp3[7:0];



              data_out[7:0] <= mem1[addr];

       end
  end


//---------------------------------------------------
always @(posedge nCE)
  begin
     nCErise = $time;
  end
  
always @(posedge nWE)
  begin
     nWErise = $time;
  end
  
wire   nWrite;  
assign nWrite = nCE || nWE;

always @(posedge nWrite)
  begin

     if ($time - adr_chng < Tsuanwelh)
       begin
           mem1[addr] <= 8'hxx;

         $display($time, "ns !!!ERROR!!!  Adr valid to end of write is not enough Tsuanwelh");
	 $finish;
       end

     if (nCErise - nCEfall < Twnce)
       begin

           mem1[addr] <= 8'hxx;

         $display($time, "ns !!!ERROR!!!  nCE to end of write is not enough Twnce");
	 $finish;
       end

     if (nCErise - da_chng < Tsudnwelh)
       begin

           mem1[addr] <= 8'hxx;

         $display($time, "ns !!!ERROR!!!  Data setup is not enough_1");
	 $finish;
       end

     if (nWErise - da_chng < Tsudnwelh)
       begin

           mem1[addr] <= 8'hxx;

         $display($time, "ns !!!ERROR!!!  Data setup is not enough_2");
	 $finish;
       end
     if (nWErise - adr_chng < Tsuanwelh)
       begin

           mem1[addr] <= 8'hxx;

         $display($time, "ns !!!ERROR!!!  Addr setup is not enough");
	 $finish;
       end
    
   if (nOE==0)
    begin
     if (nWErise - nWEfall < (Tplznwed+Tsudnwelh) )
       begin

           mem1[addr] <= 8'hxx;

         $display($time, "ns !!!ERROR!!!  WE pulse width needs to be Tplznwed+Tsudnwelh");
	 $finish;
       end

               outen  <= 1'b1;
         #(Tpzlnwed)  out_en <= outen;
    end
   if (nOE==1)
    begin
     if (nWErise - nWEfall < Twnwe)
       begin

           mem1[addr] <= 8'hxx;

         $display($time, "ns !!!ERROR!!!  WE pulse width needs to be Twnwe");
	 $finish;
       end
    end
  end

always @ (D)
  begin

     if (nWE==0 & nCE==0)
       begin

          #(Tsudnwelh) mem1[addr] <=  data_in[7:0];



              data_out[7:0] <= mem1[addr];

       end
  end
  
always @ (D) 
  begin
  da_chng <= $time;
  end
 
always @(addr)
  begin
     adr_chng <= $time;
  end

endmodule
